#include "GLWidget.h"

GLWidget::GLWidget(QGLFormat glFormat, QWidget *parent) : QGLWidget(glFormat, parent)
{
	this->_init = false;
	this->_ssbo_atm_list = nullptr;
	this->_ssbo_atm_radii = nullptr;

	_shader = nullptr;
	_computeShader = nullptr;
	_texture = nullptr;
	_grid = nullptr;
}

GLWidget::~GLWidget()
{
	cleanup();
	if (_shader != nullptr) delete _shader;
	if (_computeShader != nullptr) delete _computeShader;
	if (_texture != nullptr) delete _texture;
	if (_grid != nullptr) delete _grid;
	if (_vao != nullptr) delete _vao;
	
}

void GLWidget::cleanup()
{
	makeCurrent();

	// anything to clean up?
	if (_ssbo_atm_list != nullptr) delete _ssbo_atm_list;
	if (_ssbo_atm_radii != nullptr) delete _ssbo_atm_radii;

	doneCurrent();
}

QSize GLWidget::minimumSizeHint() const
{
	return QSize(50, 50);
}

QSize GLWidget::sizeHint() const
{
	return QSize(400, 400);
}

void GLWidget::setPDBData(std::vector<glm::vec4> atoms, std::vector<std::string> elements, std::vector<float> radii)
{
	float maxSolventRadius = 1.2f;

	this->atm_list = atoms;
	this->atm_elements = elements;
	this->atm_radii = radii;

	// get bounds
	glm::vec3 minV(FLT_MAX);
	glm::vec3 maxV(-FLT_MAX);
	for (size_t i = 0; i < atm_list.size(); i++) {
		if (atm_list[i].x < minV.x) {
			minV.x = atm_list[i].x;
		}
		if (atm_list[i].y < minV.y) {
			minV.y = atm_list[i].y;
		}
		if (atm_list[i].z < minV.z) {
			minV.z = atm_list[i].z;
		}
		if (atm_list[i].x > maxV.x) {
			maxV.x = atm_list[i].x;
		}
		if (atm_list[i].y > maxV.y) {
			maxV.y = atm_list[i].y;
		}
		if (atm_list[i].z > maxV.z) {
			maxV.z = atm_list[i].z;
		}
	}

	float maxRadius = 0.0f;
	for (size_t i = 0; i < atm_radii.size(); i++) {
		if (atm_radii[i] > maxRadius) {
			maxRadius = atm_radii[i];
		}
	}

	// extend bounds by max radius --> is done in shader
	minV = minV - glm::vec3(maxRadius * 2 + maxSolventRadius * 2);
	maxV = maxV + glm::vec3(maxRadius * 2 + maxSolventRadius * 2);

	_origin = (maxV + minV) * 0.5f;
	_angles = glm::vec2(0.0);

	_grid = new Grid(atoms, maxSolventRadius, maxRadius);


	cleanup();

	makeCurrent();

	//_vao = new VAO();
	//_vao->bind();
	_ssbo_atm_list = new SSBO(VBO::BUFFER_TYPE::POSITIONS, atm_list.size(), &atm_list[0]);
	_ssbo_atm_radii = new SSBO(VBO::BUFFER_TYPE::SCALAR, atm_radii.size(), &atm_radii[0]);
	


	_shader->setUniform("aabb.minV", minV);
	_shader->setUniform("aabb.maxV", maxV);

	_computeShader->setUniform("aabb.minV", minV);
	_computeShader->setUniform("aabb.maxV", maxV);

	_computeShader->setUniform("eps", 0.05f);
	_computeShader->setUniform("imageSize", glm::vec2(512, 512));
	_computeShader->setUniform("gridSize", _grid->getGridSize());
	_computeShader->setUniform("maxRadius", maxRadius + maxSolventRadius); // maxRadius + max solvent radius
	_computeShader->setUniform("solventRadius", maxSolventRadius); // maxRadius + max solvent radius

	doneCurrent();
	
	this->_init = true;
}

void GLWidget::initializeGL()
{
	// Set glew experimental
	glewExperimental = true;

	// Initialize GLEW
	GLenum err = glewInit();

	// If GLEW hasn't initialized
	if (err != GLEW_OK)
	{
		std::cout << "Failed to init glew" << std::endl;
	}
	
	glClearColor(0, 0, 0, 1);	

	// load shaders
	_shader = new Shader("res/shader/cast", false);
	_shader->bind();

	_computeShader = new Shader("res/shader/cast", true);
	_texture = new Texture2D(512, 512, TEXTURE_FORMAT::RGBA32F, TEXTURE_COMPONENTS::RGB, TEXTURE_DATA_TYPE::FLOAT, SAMPLING_MODE::LINEAR, SAMPLER_WRAP_MODE::CLAMP_TO_BORDER, false);


	_vao = new VAO();
	_vao->bind();
}

void GLWidget::resizeGL(int width, int height)
{
	_windowWidth = width;
	_windowHeight = height;
	glViewport(0, 0, width, height);
}

void GLWidget::paintGL() 
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (!this->_init) return;

	
	_computeShader->bind();

	_texture->bindImage(0);
	_computeShader->setUniform("depthTex", 0);

	glm::vec4 camPos = glm::vec4(0.0, 0.0, 1.0, 1.0);
	camPos = glm::rotateY(camPos, glm::radians(_angles.x));

	_computeShader->setUniform("dir", glm::vec3(camPos));
	_computeShader->setUniform("origin", _origin);

	_ssbo_atm_list->bind();
	_ssbo_atm_list->bindBase(0);
	_ssbo_atm_radii->bind();
	_ssbo_atm_radii->bindBase(1);
	SSBO* gridData = _grid->getSSBOGridData();
	gridData->bind();
	gridData->bindBase(2);

	glDispatchCompute(512, 1, 1);

	// block until compute shader finishes, otherwise render call would result with strange results
	// block for rendercall GL_VERTEX_ATTRIB_ARRAY_BARRIER_BIT and for next dispatch GL_SHADER_STORAGE_BARRIER_BIT
	glMemoryBarrier(GL_ALL_BARRIER_BITS);


	_shader->bind();
	_shader->setUniform("depthTex", _texture, 0);
	_shader->setUniform("dir", glm::vec3(camPos));

	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

	GLenum err;
	while ((err = glGetError()) != GL_NO_ERROR) {
		std::cout << "OpenGL error: " << err << std::endl;
	}
}

void GLWidget::mouseMoveEvent(QMouseEvent *eventMove)
{
	if (_rotating) {
		int newX = eventMove->x();
		int newY = eventMove->y();

		if (newX != _oldX || newY != _oldY) {
			int dx = newX - _oldX;
			int dy = newY - _oldY;

			float speed = 0.5f;

			_angles.x += dx * speed;
			_angles.y += dy * speed;

			_oldX = newX;
			_oldY = newY;
		}
	}
}

void GLWidget::mousePressEvent(QMouseEvent *eventPress)
{
	if (eventPress->button() == Qt::LeftButton) {
		_rotating = true;
		_oldX = eventPress->x();
		_oldY = eventPress->y();
	}
}

void GLWidget::mouseReleaseEvent(QMouseEvent *releaseEvent)
{
	_rotating = false;
}
