﻿#include "CubeMapTexture.h"


CubeMapTexture::CubeMapTexture(	unsigned int width, unsigned int height,
								const unsigned char* face_x_pos, const unsigned char* face_x_neg,
								const unsigned char* face_y_pos, const unsigned char* face_y_neg,
								const unsigned char* face_z_pos, const unsigned char* face_z_neg,
								TEXTURE_FORMAT format, TEXTURE_COMPONENTS components, TEXTURE_DATA_TYPE datatype,
								SAMPLING_MODE sampling, SAMPLER_WRAP_MODE wrap, bool srgb)
	: _width(width), _height(height), _srgb(srgb)
{
	glGenTextures(1, &_handle);

	GLenum texTarget = GL_TEXTURE_CUBE_MAP;

	unsigned int format_ = (unsigned int)format;
	if (srgb) {
		switch (format)
		{
		case TEXTURE_FORMAT::RGB8:
			format_ = GL_SRGB8;
			break;
		case TEXTURE_FORMAT::RGBA8:
			format_ = GL_SRGB8_ALPHA8;
			break;
		}
	}

	glBindTexture(texTarget, _handle);

	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X, 0, format_, _width, _height, 0, (unsigned int)components, (unsigned int)datatype, face_x_pos);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_X, 0, format_, _width, _height, 0, (unsigned int)components, (unsigned int)datatype, face_x_neg);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Y, 0, format_, _width, _height, 0, (unsigned int)components, (unsigned int)datatype, face_y_pos);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, 0, format_, _width, _height, 0, (unsigned int)components, (unsigned int)datatype, face_y_neg);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_Z, 0, format_, _width, _height, 0, (unsigned int)components, (unsigned int)datatype, face_z_pos);
	glTexImage2D(GL_TEXTURE_CUBE_MAP_NEGATIVE_Z, 0, format_, _width, _height, 0, (unsigned int)components, (unsigned int)datatype, face_z_neg);

	if (components == TEXTURE_COMPONENTS::DEPTH) {
		// not implemented yet
	}
	else {
		glTexParameteri(texTarget, GL_TEXTURE_WRAP_S, (unsigned int)wrap);
		glTexParameteri(texTarget, GL_TEXTURE_WRAP_T, (unsigned int)wrap);
		glTexParameteri(texTarget, GL_TEXTURE_WRAP_R, (unsigned int)wrap);
		switch (sampling)
		{
		case SAMPLING_MODE::LINEAR:
			glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			break;
		case SAMPLING_MODE::BILINEAR:
			glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			break;
		case SAMPLING_MODE::TRILINEAR:
			glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			break;
		}
	}

	glGenerateMipmap(GL_TEXTURE_CUBE_MAP);

	glEnable(GL_TEXTURE_CUBE_MAP_SEAMLESS);

	unbind();
}

CubeMapTexture::~CubeMapTexture()
{
	glDeleteTextures(1, &_handle);
}

void CubeMapTexture::getTextureDimensions(int& width, int& height) const
{
	width = _width;
	height = _height;
}

void CubeMapTexture::bind(unsigned int unit) const
{
	glActiveTexture(GL_TEXTURE0 + unit);
	glBindTexture(GL_TEXTURE_CUBE_MAP, _handle);
}

void CubeMapTexture::unbind() const
{
	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
}

unsigned int CubeMapTexture::getHandle() const
{
	return _handle;
}

void CubeMapTexture::generateMipMaps() const
{
	glBindTexture(GL_TEXTURE_CUBE_MAP, _handle);
	glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
}

bool CubeMapTexture::isSRGBImage() const
{
	return _srgb;
}