#include "Shader.h"


Shader::Shader(std::string file, bool computeShader) 
	: _file(file), _computeShader(computeShader)
{
	_handle = loadShaders(computeShader);
}

Shader::~Shader()
{
	glUseProgram(0);
	glDeleteProgram(_handle);
}

GLuint Shader::loadShaders(bool computeShader)
{
	GLuint handle = 0;
	if (!computeShader) {
		GLuint vs, fs, gms, tcs, tes;
		bool bvs, bfs, bgms, btcs, btes;
		bvs = loadShader_(_file, GL_VERTEX_SHADER, EXT_VERTEX_SHADER, vs);
		bfs = loadShader_(_file, GL_FRAGMENT_SHADER, EXT_FRAGMENT_SHADER, fs);
		bgms = loadShader_(_file, GL_GEOMETRY_SHADER, EXT_GEOMETRY_SHADER, gms);
		btcs = loadShader_(_file, GL_TESS_CONTROL_SHADER, EXT_TESSELATION_CONTROL_SHADER, tcs);
		btes = loadShader_(_file, GL_TESS_EVALUATION_SHADER, EXT_TESSELATION_EVALUATION_SHADER, tes);

		if (!bvs && !bfs) {
			std::cout << "Shader must contain at least a vertex and fragment shader" << std::endl;
			system("PAUSE");
			exit(EXIT_FAILURE);
		}

		handle = glCreateProgram();

		// Attach shader
		if (bvs)
			glAttachShader(handle, vs);
		if (bfs)
			glAttachShader(handle, fs);
		if (bgms)
			glAttachShader(handle, gms);
		if (btcs)
			glAttachShader(handle, tcs);
		if (btes)
			glAttachShader(handle, tes);

		glLinkProgram(handle);

		// Check for errors
		GLint succeded = 0;
		glGetProgramiv(handle, GL_LINK_STATUS, &succeded);

		if (succeded == GL_FALSE) {
			// Print error
			GLint logSize;
			glGetProgramiv(handle, GL_INFO_LOG_LENGTH, &logSize);

			char* message = new char[logSize];
			glGetProgramInfoLog(handle, logSize, nullptr, message);

			std::cout << "Failed to link shader program:" << std::endl;
			std::cout << message << std::endl;



			// Delete shader & programm
			glDeleteProgram(handle);
			if (bvs)
				glDeleteShader(vs);
			if (bfs)
				glDeleteShader(fs);
			if (bgms)
				glDeleteShader(gms);
			if (btcs)
				glDeleteShader(tcs);
			if (btes)
				glDeleteShader(tes);

			system("PAUSE");
			delete[] message;
			exit(EXIT_FAILURE);
		}

		// After Link detach and delete Shader
		if (bvs)
			glDetachShader(handle, vs);
		if (bfs)
			glDetachShader(handle, fs);
		if (bgms)
			glDetachShader(handle, gms);
		if (btcs)
			glDetachShader(handle, tcs);
		if (btes)
			glDetachShader(handle, tes);

		if (bvs)
			glDeleteShader(vs);
		if (bfs)
			glDeleteShader(fs);
		if (bgms)
			glDeleteShader(gms);
		if (btcs)
			glDeleteShader(tcs);
		if (btes)
			glDeleteShader(tes);
	} else {
		GLuint cs;
		bool bcs;

		bcs = loadShader_(_file, GL_COMPUTE_SHADER, EXT_COMPUTE_SHADER, cs);

		if (!bcs) {
			std::cout << "Failed to compile compute shader" << std::endl;
			system("PAUSE");
			exit(EXIT_FAILURE);
		}

		handle = glCreateProgram();
		glAttachShader(handle, cs);
		glLinkProgram(handle);

		// Check for errors
		GLint succeded = 0;
		glGetProgramiv(handle, GL_LINK_STATUS, &succeded);

		if (succeded == GL_FALSE) {
			// Print error
			GLint logSize;
			glGetProgramiv(handle, GL_INFO_LOG_LENGTH, &logSize);

			char* message = new char[logSize];
			glGetProgramInfoLog(handle, logSize, nullptr, message);

			std::cout << "Failed to link compute shader program:" << std::endl;
			std::cout << message << std::endl;

			// Delete shader & programm
			glDeleteProgram(handle);
			glDeleteShader(cs);

			system("PAUSE");
			delete[] message;
			exit(EXIT_FAILURE);
		}

		glDetachShader(handle, cs);
		glDeleteShader(cs);
	}

	return handle;
}

bool Shader::loadShader_(std::string file, GLenum shaderType, std::string ext, GLuint& handle)
{
	//get Stream
	std::ifstream shaderFile(file + ext);

	if (shaderFile.good()) {
		//Read  Shader code from file
		std::string shaderCode = std::string(std::istreambuf_iterator<char>(shaderFile), std::istreambuf_iterator<char>());
		shaderFile.close();

		//create shader handle
		handle = glCreateShader(shaderType);

		//If Handle is 0 there was an Error
		if (handle == 0){
			std::cout << "Failed to create " << shaderType << std::endl;
			system("PAUSE");
			exit(EXIT_FAILURE);
		}

		//Compile Shader
		const char* codePtr = shaderCode.c_str();
		glShaderSource(handle, 1, &codePtr, 0);
		glCompileShader(handle);

		//Test for compilation errors
		GLint succeded;
		glGetShaderiv(handle, GL_COMPILE_STATUS, &succeded);

		if (succeded == GL_FALSE || !glIsShader(handle)){
			//Read Log and output it to cout
			GLint logSize;
			glGetShaderiv(handle, GL_INFO_LOG_LENGTH, &logSize);

			char* message = new char[logSize];
			glGetShaderInfoLog(handle, logSize, nullptr, message);

			std::cout << "Failed to compile shader: " << _file << std::endl;
			std::cout << message << std::endl;

			//delete shader & programm
			glDeleteShader(handle);

			system("PAUSE");
			delete[] message;
			exit(EXIT_FAILURE);
		}
	}
	else{
		return false;
	}

	return true;
}

void Shader::bind() 
{
	glUseProgram(_handle);
}

void Shader::unbind()
{
	glUseProgram(0);
}

void Shader::reload()
{
	glUseProgram(0);
	glDeleteProgram(_handle);
	_handle = loadShaders(_computeShader);
}

void Shader::setUniform(const char* name, int i)
{
	setUniform(glGetUniformLocation(_handle, name), i);
}

void Shader::setUniform(GLint location, int i)
{
	if (location != -1)
		glProgramUniform1i(_handle, location, i);
}

void Shader::setUniform(const char* name, unsigned int i)
{
	setUniform(glGetUniformLocation(_handle, name), i);
}

void Shader::setUniform(GLint location, unsigned int i)
{
	if (location != -1)
		glProgramUniform1ui(_handle, location, i);
}

void Shader::setUniform(const char* name, float f)
{
	setUniform(glGetUniformLocation(_handle, name), f);
}

void Shader::setUniform(GLint location, float f)
{
	if (location != -1)
		glProgramUniform1f(_handle, location, f);
}

void Shader::setUniform(const char* name, const glm::mat4& mat)
{
	setUniform(glGetUniformLocation(_handle, name), mat);
}

void Shader::setUniform(GLint location, const glm::mat4& mat)
{
	if (location != -1)
		glProgramUniformMatrix4fv(_handle, location, 1, GL_FALSE, glm::value_ptr(mat));
}

void Shader::setUniform(const char* name, const glm::mat3& mat)
{
	setUniform(glGetUniformLocation(_handle, name), mat);
}

void Shader::setUniform(GLint location, const glm::mat3& mat)
{
	if (location != -1)
		glProgramUniformMatrix3fv(_handle, location, 1, GL_FALSE, glm::value_ptr(mat));
}

void Shader::setUniform(const char* name, const glm::vec2& vec)
{
	setUniform(glGetUniformLocation(_handle, name), vec);
}

void Shader::setUniform(GLint location, const glm::vec2& vec)
{
	if (location != -1)
		glProgramUniform2f(_handle, location, vec.x, vec.y);
}

void Shader::setUniform(const char* name, const glm::vec3& vec)
{
	setUniform(glGetUniformLocation(_handle, name), vec);
}

void Shader::setUniform(GLint location, const glm::vec3& vec)
{
	if (location != -1)
		glProgramUniform3f(_handle, location, vec.x, vec.y, vec.z);
}

void Shader::setUniform(const char* name, Light* l)
{
	std::string prefix = std::string(name) + ".";
	if (l->type == LIGHT_TYPE::DIRECTIONAL) {
		setUniform((prefix + "direction").c_str(), l->direction);
	}
	else if (l->type == LIGHT_TYPE::POINT) {
		setUniform((prefix + "position").c_str(), l->position);
		setUniform((prefix + "attenuation").c_str(), l->attenuation);
	}
	setUniform((prefix + "diffColor").c_str(), l->colorDiffuse);
	setUniform((prefix + "specColor").c_str(), l->colorSpecular);
}

void Shader::setUniformArr(const char* name, Light* l, int index)
{
	setUniform((std::string(name) + "[" + std::to_string(index) + "]").c_str(), l);
}

void Shader::setUniformArr(const char* name, const glm::vec2* arr, int index)
{
	setUniform(name, arr[index]);
}

void Shader::setUniform(const char* name, const glm::vec4& vec)
{
	setUniform(glGetUniformLocation(_handle, name), vec);
}

void Shader::setUniform(GLint location, const glm::vec4& vec)
{
	if (location != -1)
		glProgramUniform4f(_handle, location, vec.x, vec.y, vec.z, vec.w);
}

void Shader::setUniform(const char* name, const Texture2D* tex, unsigned int unit)
{
	GLint location = glGetUniformLocation(_handle, name);
	if (location != -1) {
		tex->bind(unit);
		glProgramUniform1i(_handle, location, unit);
	}
}

void Shader::setUniform(const char* name, const CubeMapTexture* tex, unsigned int unit)
{
	GLint location = glGetUniformLocation(_handle, name);
	if (location != -1) {
		tex->bind(unit);
		glProgramUniform1i(_handle, location, unit);
	}
}

void Shader::setUniform(const char* name, const Texture3D* tex, unsigned int unit)
{
	GLint location = glGetUniformLocation(_handle, name);
	if (location != -1) {
		tex->bind(unit);
		glProgramUniform1i(_handle, location, unit);
	}
}

void Shader::setUniform(const char* name, const glm::mat4x2& mat)
{
	setUniform(glGetUniformLocation(_handle, name), mat);
}

void Shader::setUniform(GLint location, const glm::mat4x2& mat)
{
	if (location != -1)
		glUniformMatrix4x2fv(location, 1, GL_FALSE, glm::value_ptr(mat));
}