#include "Texture2D.h"

void Texture2D::initTex2D(TEXTURE_FORMAT format, TEXTURE_COMPONENTS components, TEXTURE_DATA_TYPE datatype, SAMPLING_MODE sampling, SAMPLER_WRAP_MODE wrap, const void* data)
{
	glGenTextures(1, &_handle);

	GLenum texTarget = GL_TEXTURE_2D;

	_format = format;
	unsigned int format_ = (unsigned int)format;
	if (_srgb) {
		switch (format)
		{
			case TEXTURE_FORMAT::RGB8:
				format_ = GL_SRGB8;
				break;
			case TEXTURE_FORMAT::RGBA8:
				format_ = GL_SRGB8_ALPHA8;
				break;
		}
	}

	glBindTexture(texTarget, _handle);
	
	glTexImage2D(texTarget, 0, format_, _width, _height, 0, (unsigned int)components, (unsigned int)datatype, data);

	if (components == TEXTURE_COMPONENTS::DEPTH) {
		glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(texTarget, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
		glTexParameteri(texTarget, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
		glTexParameteri(texTarget, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_REF_TO_TEXTURE);
		glTexParameteri(texTarget, GL_TEXTURE_COMPARE_FUNC, GL_LESS);
		float ones[] = { 1.0f, 1.0f, 1.0f, 1.0f };
		glTexParameterfv(texTarget, GL_TEXTURE_BORDER_COLOR, ones);
	} else {
		glTexParameteri(texTarget, GL_TEXTURE_WRAP_S, (unsigned int)wrap);
		glTexParameteri(texTarget, GL_TEXTURE_WRAP_T, (unsigned int)wrap);
		switch (sampling)
		{
		case SAMPLING_MODE::LINEAR:
			glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			break;
		case SAMPLING_MODE::BILINEAR:
			glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			break;
		case SAMPLING_MODE::TRILINEAR:
			glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			break;
		}

		if (data != nullptr) {
			glGenerateMipmap(texTarget);
		}
	}

	unbind();
}

Texture2D::Texture2D(unsigned int width, unsigned int height, TEXTURE_FORMAT format, TEXTURE_COMPONENTS components, TEXTURE_DATA_TYPE datatype, SAMPLING_MODE sampling, SAMPLER_WRAP_MODE wrap, bool srgb)
	: Texture2D(nullptr, width, height, format, components, datatype, sampling, wrap, srgb)
{
}

Texture2D::Texture2D(const void * data, unsigned int width, unsigned int height, TEXTURE_FORMAT format, TEXTURE_COMPONENTS components, TEXTURE_DATA_TYPE datatype, SAMPLING_MODE sampling, SAMPLER_WRAP_MODE wrap, bool srgb)
	: _width(width), _height(height), _srgb(srgb)
{
	initTex2D(format, components, datatype, sampling, wrap, data);
}

Texture2D::~Texture2D()
{
	glDeleteTextures(1, &_handle);
}

void Texture2D::getTextureDimensions(int& width, int& height) const
{
	width = _width;
	height = _height;
}

void Texture2D::bind(unsigned int unit) const
{
	GLenum texTarget = GL_TEXTURE_2D;

	glActiveTexture(GL_TEXTURE0 + unit);
	glBindTexture(texTarget, _handle);
}

void Texture2D::bindImage(unsigned int unit) const
{
	glActiveTexture(GL_TEXTURE0 + unit);
	GLenum texTarget = GL_TEXTURE_2D;
	glBindTexture(texTarget, _handle);
	glBindImageTexture(unit, _handle, 0, GL_FALSE, 0, GL_WRITE_ONLY, (unsigned int)_format);
}

void Texture2D::unbind() const
{
	GLenum texTarget = GL_TEXTURE_2D;

	glBindTexture(texTarget, 0);
}

unsigned int Texture2D::getHandle() const
{
	return _handle;
}

void Texture2D::setSamplingMode(SAMPLING_MODE samplingmode)
{
	switch (samplingmode)
	{
		case SAMPLING_MODE::LINEAR:
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			break;
		case SAMPLING_MODE::BILINEAR:
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			break;
		case SAMPLING_MODE::TRILINEAR:
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			break;
	}
}

void Texture2D::generateMipMaps() const
{
	GLenum texTarget = GL_TEXTURE_2D;

	glBindTexture(texTarget, _handle);
	glGenerateMipmap(texTarget);
	glBindTexture(texTarget, 0);
}

bool Texture2D::isSRGBImage() const
{
	return _srgb;
}