#include "SceneObject.h"

SceneObject::SceneObject(std::string name) : SceneObject(name, glm::mat4(1))
{
}

SceneObject::SceneObject(std::string name, const glm::mat4& model)
	: SceneObject(name, model, nullptr)
{
}

SceneObject::SceneObject(std::string name, const glm::mat4& model, SceneObject* parent)
	: _name(name), _model(model), _inverseModel(glm::inverse(model)), _parent(parent), _enabled(true)
{

	bool disableViewFrustumCulling = false;
	if (_parent != nullptr)
		parent->addChild(this);
}

SceneObject::~SceneObject()
{
	bool disableViewFrustumCulling = false;
	for (size_t i = 0; i < _meshes.size(); i++) {
		if (_meshes[i] != nullptr) {
			delete _meshes[i];
		}
	}

	for (size_t i = 0; i < _children.size(); i++) {
		delete _children[i];
	}
}

glm::mat4 SceneObject::getModelMatrix() const
{
	return _model;
}

glm::mat4 SceneObject::getInverseModelMatrix() const
{
	return _inverseModel;
}

void SceneObject::addMesh(Mesh * mesh)
{
	_meshes.push_back(mesh);
}

void SceneObject::addChild(SceneObject* child)
{
	_children.push_back(child);
}

void SceneObject::update(float dt)
{
	for (size_t i = 0; i < _children.size(); i++) {
		_children[i]->update(dt);
	}

	for (size_t i = 0; i < _meshes.size(); i++) {
		_meshes[i]->update(dt);
	}

	postUpdate(dt);
}

void SceneObject::postUpdate(float dt)
{
	// empty
}

SceneObject* SceneObject::find(std::string name)
{
	if (_name == name) return this;
	
	SceneObject* obj = nullptr;

	for (unsigned int i = 0; i < _children.size(); i++) {
		obj = _children[i]->find(name);
		if (obj != nullptr) return obj;
	}

	return obj;
}

void SceneObject::render(RenderInfo& renderinfo, const glm::mat4& Maccum, const glm::mat4& V, const glm::mat4& P) const
{
	if (!_enabled) return;

	glm::mat4 M = Maccum * _model;

	for (size_t i = 0; i < _meshes.size(); i++) {
		if (_meshes[i] != nullptr) {
			_meshes[i]->render(renderinfo, M, V, P);
		}
	}

	for (size_t i = 0; i < _children.size(); i++) {
		_children[i]->render(renderinfo, M, V, P);
	}	
}

AABB SceneObject::getAABB() const
{
	std::vector<AABB> aabbs;
	for (unsigned int i = 0; i < _meshes.size(); i++) {
		aabbs.push_back(_meshes[i]->getAABB());
	}
	for (unsigned int i = 0; i < _children.size(); i++) {
		aabbs.push_back(_children[i]->getAABB());
	}

	return AABB(aabbs);
}

void SceneObject::initAABB(glm::mat4 M)
{
	initAABBRec(M);
}

void SceneObject::initAABBRec(const glm::mat4 & Maccum)
{
	glm::mat4 M = Maccum * _model;
	for (unsigned int i = 0; i < _meshes.size(); i++) {
		AABB ws_aabb = _meshes[i]->getAABB().transform(M);
		std::vector<glm::vec3> box_vert(8);
		for (int i = 0; i < 8; i++) {
			box_vert.push_back(ws_aabb.box[i]);
		}
		_meshes[i]->setAABB(AABB(box_vert));
	}

	for (unsigned int i = 0; i < _children.size(); i++) {
		_children[i]->initAABBRec(M);
	}
}
